import pytest

from matterlab_valves import JKem4in1Valve


@pytest.mark.real
def test_jkem_4in1_real_read(hardware_com_port, hardware_valve_num):
    valve = JKem4in1Valve(
        com_port=hardware_com_port,
        valve_num=hardware_valve_num,
        ports={"1": 1, "2": 2},
    )

    current_port = valve.port

    assert isinstance(current_port, int)
    assert 1 <= current_port <= valve.num_port


@pytest.mark.real
def test_jkem_4in1_real_switch(hardware_com_port, hardware_valve_num):
    valve = JKem4in1Valve(
        com_port=hardware_com_port,
        valve_num=hardware_valve_num,
        ports={"1": 1, "2": 2},
    )

    valve.port = 1
    assert valve.port == 1

    valve.switch_port("2")
    assert valve.port == 2



from matterlab_valves import RunzeSelectionValve, create_runze_selection_sim_transport_factory
from matterlab_valves.runze_valve import RunzeMotorEventType


def test_sim_transport_reports_initial_port():
    transport_factory = create_runze_selection_sim_transport_factory(address=1, num_port=10, initial_port=2)
    valve = RunzeSelectionValve(com_port="SIM::runze-valve", address=1, num_port=10, transport_factory=transport_factory)

    assert valve.port == 2


def test_sim_transport_moves_and_homes_with_public_api():
    transport_factory = create_runze_selection_sim_transport_factory(address=1, num_port=10)
    valve = RunzeSelectionValve(
        com_port="SIM::runze-valve",
        address=1,
        num_port=10,
        ports={"home": 1, "collect": 2, "waste": 3},
        transport_factory=transport_factory,
    )

    valve.port = 3
    assert valve.port == 3

    valve.switch_port("home")
    assert valve.port == 1
    assert transport_factory.state.current_port == 1


def test_sim_transport_check_motor_uses_simulated_state():
    transport_factory = create_runze_selection_sim_transport_factory(
        address=1,
        num_port=10,
        motor_status=RunzeMotorEventType.Normal_status,
    )
    valve = RunzeSelectionValve(com_port="SIM::runze-valve", address=1, num_port=10, transport_factory=transport_factory)

    assert valve.check_motor() == RunzeMotorEventType.Normal_status.value


def test_sim_transport_reset_returns_home():
    transport_factory = create_runze_selection_sim_transport_factory(address=1, num_port=10, initial_port=4)
    valve = RunzeSelectionValve(com_port="SIM::runze-valve", address=1, num_port=10, transport_factory=transport_factory)

    valve.reset()

    assert valve.port == 1
    assert transport_factory.state.current_port == 1


def test_sim_transport_between_ports_tracks_state():
    transport_factory = create_runze_selection_sim_transport_factory(address=1, num_port=10, initial_port=4)
    valve = RunzeSelectionValve(com_port="SIM::runze-valve", address=1, num_port=10, transport_factory=transport_factory)

    valve.set_between_ports(4)

    assert transport_factory.state.between_ports_after == 4



from pathlib import Path
import time
from logging import Logger
from typing import Dict, Tuple, Optional

from matterlab_serial_device import SerialDevice, open_close

from matterlab_valves.base_valve import SwitchingValve


class JKem4in1Valve(SwitchingValve, SerialDevice):
    category="Valve"
    ui_fields  = ("com_port", "valve_num", "num_port")
    def __init__(self,
                 com_port: str,
                 valve_num: int,
                 num_port: int = 8,
                 encoding: str = "utf-8",
                 baudrate: int = 9600,
                 timeout: float = 1.0,
                 parity: str = "none",
                 bytesize: int = 8,
                 stopbits: int = 1,
                 ports: Optional[Dict[str, int]] = None,
                 connect_hardware: bool = True,
                 logger: Optional[Logger] = None,
                 **kwargs
                 ):
        """
        :param com_port: COM port of the pump connected to, example: "COM1", "/tty/USB0"
        :param ports: dict containing nickname and valve port number mapping
        :param connect_hardware: if connect hardware up on initialization
        :return:
        """

        SerialDevice.__init__(self,
                              com_port=com_port,
                              encoding=encoding,
                              baudrate=baudrate,
                              timeout=timeout,
                              parity=parity,
                              bytesize=bytesize,
                              stopbits=stopbits,
                              **kwargs
                              )
        SwitchingValve.__init__(self,
                                num_port=num_port,
                                ports=ports,
                                logger=logger
                                )
        self.valve_num = valve_num
        if connect_hardware:
            self.connect()

    def connect(self):
        """
        placeholder to homogenize hardware initialization
        keep here to be consistent with other hardware code
        :return:
        """
        pass

    def _verify_valve_num(self):
        if hasattr(self.settings, 'valve_num'):
            if self.valve_num in (1, 2, 3, 4):
                return
            else:
                raise ValueError("Valve_num must be in 1-4.")
        else:
            self.logger.warning(
                "Using port property for JKem valve requires specifying valve_num."
            )
        #raise ValueError("JKem need to specify valve_num when instantiating.")

    @open_close
    def _execute_valve(self, valve_num: int, port: int):
        """
        execute operations on valve and port
        """
        self.write(command=f'/{valve_num}o{port}R\r')

    @open_close
    def _query_valve(self, valve_num: int) -> str:
        return  self.query(write_command=f'/{valve_num}?8\r',
                           read_until='\r',
                           remove_from_start=3,
                           remove_from_end=1)

    def get_current_port(self, valve_num: int, max_try: int = 5) -> int:
        for i in range(0, max_try):
            try:
                port = int(self._query_valve(valve_num = valve_num), base = 10)
                self.logger.debug("JKem 4-in-1 valve on %s No. %s reports current port %s", self.com_port, valve_num, port)
                return port
            except:
                time.sleep(1)
                continue
        raise IOError(f'Query port of valve {self.com_port} No. {valve_num} exceed max try.')

    def set_current_port(self, valve_num: int, port: int, max_try: int = 5):
        self.logger.info("Requesting JKem 4-in-1 valve move on %s No. %s to port %s", self.com_port, valve_num, port)
        for i in range(0, max_try):
            self._execute_valve(valve_num = valve_num, port = port)
            if self.get_current_port(valve_num = valve_num) == port:
                self.logger.info(
                    "Valve on %s No. %s has been moved to %s",
                    self.com_port,
                    valve_num,
                    port,
                )
                time.sleep(1)
                return None
            else:
                time.sleep(1)
                continue
        raise IOError(f'Move port of valve {self.com_port} No. {valve_num} exceed max try.')

    @property
    def port(self) -> int:
        """
        concrete method to get current port
        """
        return self.get_current_port(valve_num=self.valve_num)

    @port.setter
    def port(self, port: int):
        """
        concrete method to set current port
        """
        self.set_current_port(valve_num=self.valve_num, port = port)






import pytest

from matterlab_valves import SwitchingValve


class MockSwitchingValve(SwitchingValve):
    def __init__(self, num_port, ports=None):
        super().__init__(num_port=num_port, ports=ports)
        self._port = 1

    @property
    def port(self):
        return self._port

    @port.setter
    def port(self, port_id):
        self._port = port_id


def test_switching_valve_requires_concrete_subclass():
    with pytest.raises(TypeError):
        SwitchingValve(num_port=2, ports={"A": 1, "B": 2})


def test_switching_valve_stores_num_port_and_ports():
    ports = {"A": 1, "B": 2, "C": 3}
    valve = MockSwitchingValve(num_port=3, ports=ports)

    assert isinstance(valve, SwitchingValve)
    assert valve.num_port == 3
    assert valve.ports == ports


def test_switch_port_accepts_named_port():
    valve = MockSwitchingValve(num_port=2, ports={"A": 1, "B": 2})

    result = valve.switch_port("B")

    assert result == 2
    assert valve.port == 2


def test_switch_port_accepts_numeric_port():
    valve = MockSwitchingValve(num_port=3, ports={"A": 1, "B": 2, "C": 3})

    result = valve.switch_port(3)

    assert result == 3
    assert valve.port == 3


def test_switch_port_raises_for_unknown_port():
    valve = MockSwitchingValve(num_port=2, ports={"A": 1, "B": 2})

    with pytest.raises(ValueError, match="Wrong ports"):
        valve.switch_port("C")


def test_get_port_id_returns_matching_label():
    valve = MockSwitchingValve(num_port=2, ports={"A": 1, "B": 2})

    assert valve._get_port_id(1) == "A"
    assert valve._get_port_id(2) == "B"


def test_get_port_id_raises_for_unknown_port_number():
    valve = MockSwitchingValve(num_port=2, ports={"A": 1, "B": 2})

    with pytest.raises(KeyError, match="Port number 3 not found in ports."):
        valve._get_port_id(3)


def test_port_property_uses_concrete_implementation():
    valve = MockSwitchingValve(num_port=2, ports={"A": 1, "B": 2})

    assert valve.port == 1

    valve.port = 2

    assert valve.port == 2



from multiprocessing import Value
from pathlib import Path
import time
from logging import Logger
from typing import Dict, Tuple, Optional, Union
from struct import pack, unpack

from matterlab_serial_device import SerialDevice, open_close

from matterlab_valves.base_valve import SwitchingValve


class JKemNewValve(SwitchingValve, SerialDevice):
    category="Valve"
    ui_fields  = ("com_port", "valve_num", "num_port")
    def __init__(self,
                 com_port: str,
                 valve_num: int,
                 num_port: int = 8,
                 encoding: str = "utf-8",
                 baudrate: int = 38400,
                 timeout: float = 1.0,
                 parity: str = "none",
                 bytesize: int = 8,
                 stopbits: int = 1,
                 ports: Optional[Dict[str, int]] = None,
                 connect_hardware: bool = True,
                 logger: Optional[Logger] = None,
                 **kwargs
                 ):
        """
        :param com_port: COM port of the pump connected to, example: "COM1", "/tty/USB0"
        :param ports: dict containing nickname and valve port number mapping
        :param connect_hardware: if connect hardware up on initialization
        :return:
        """
        
        SerialDevice.__init__(self,
                              com_port=com_port,
                              encoding=encoding,
                              baudrate=baudrate,
                              timeout=timeout,
                              parity=parity,
                              bytesize=bytesize,
                              stopbits=stopbits,
                              **kwargs
                              )
        SwitchingValve.__init__(self,
                                num_port=num_port,
                                ports=ports,
                                logger=logger
                                )
        self.valve_num = valve_num
        if connect_hardware:
            self.connect(self.valve_num)
            
    @staticmethod
    def calculate_checksum(raw_cmd_byte: Union[bytes, bytearray])->Union[bytes, bytearray]:
        sum_of_bytes = 0
        for b in raw_cmd_byte:
            sum_of_bytes += b
        return raw_cmd_byte + pack("<H", sum_of_bytes)
        
    @open_close
    def connect(self, valve_num: int = 1):
        """
        connect device, raise error if failed, default try to connect valve 1 
        :return: 0 for success
        """
        self.logger.debug("Connecting JKem valve on %s No. %s", self.com_port, valve_num)
        cmd = bytearray([204, valve_num, 74, 0, 0, 221])
        cmd_checked = self.calculate_checksum(cmd)
        rtn = self.query(write_command = cmd_checked,
                   read_until="",
                   remove_from_start=2,
                   remove_from_end=5
                   )
        if len(rtn) != 1:
            raise IOError("Connecting valve failed!")
        if unpack("B", rtn)[0] != 0:
            raise IOError("Connecting valve failed!")
        self.logger.info("Connected JKem valve on %s No. %s", self.com_port, valve_num)
        return 0
    
    @open_close
    def _execute_valve(self, valve_num: int, port: int)->int:
        """
        execute operation on valve and port
        :return: -1 for failure, 0 for success
        """
        cmd = bytearray([204, valve_num, 68, port, 0, 221])
        cmd_checked = self.calculate_checksum(cmd)
        rtn = self.query(write_command = cmd_checked,
                   read_until=""
                   )
        if len(rtn) != 8:
            return -1
        else:
            return 0
        
    @open_close
    def _query_valve(self, valve_num: int) -> int:
        """
        query valve for port info
        :return: current port number, or -1 for failue
        """
        cmd = bytearray([204, valve_num, 62, 0, 0, 221])
        cmd_checked = self.calculate_checksum(cmd)
        rtn =  self.query(write_command=cmd_checked,
                          read_until="",
                          remove_from_start=3,
                          remove_from_end=4
                          )
        if len(rtn) != 1:
            return -1
        return unpack("B", rtn)[0]

    
    def get_current_port(self, valve_num: int, max_try: int = 5) -> int:
        for i in range(0, max_try):
            try:
                port = self._query_valve(valve_num = valve_num)
                self.logger.debug("JKem valve on %s No. %s reports current port %s", self.com_port, valve_num, port)
                return port
            except:
                time.sleep(1)
                continue
        raise IOError(f'Query port of valve {self.com_port} No. {valve_num} exceed max try.')

    def set_current_port(self, valve_num: int, port: int, max_try: int = 5):
        self.logger.info("Requesting JKem valve move on %s No. %s to port %s", self.com_port, valve_num, port)
        for i in range(0, max_try):
            self._execute_valve(valve_num = valve_num, port = port)
            if self.get_current_port(valve_num = valve_num) == port:
                self.logger.info(
                    "Valve on %s No. %s has been moved to %s",
                    self.com_port,
                    valve_num,
                    port,
                )
                time.sleep(1)
                return None
            else:
                time.sleep(1)
                continue
        raise IOError(f'Move port of valve {self.com_port} No. {valve_num} exceed max try.')

    @property
    def port(self) -> int:
        """
        concrete method to get current port
        """
        return self.get_current_port(valve_num=self.valve_num)

    @port.setter
    def port(self, port: int):
        """
        concrete method to set current port
        """
        self.set_current_port(valve_num=self.valve_num, port = port)



import os
from struct import pack
from unittest.mock import patch

import pytest

from matterlab_serial_device import SerialDevice
from matterlab_valves.jkem_new import JKemNewValve


PORT = "/dev/ttyUSB0" if os.name == "posix" else "COM1"


@pytest.fixture
def mock_serial(mocker):
    mock_serial_cls = mocker.patch("matterlab_serial_device.transport.serial.Serial", autospec=True)
    serial_instances = []
    primary_instance = mocker.MagicMock()
    primary_instance.com_port = PORT
    primary_instance._factory = mock_serial_cls
    primary_instance._instances = serial_instances

    def _make_serial_instance(*args, **kwargs):
        if not serial_instances:
            instance = primary_instance
        else:
            instance = mocker.MagicMock()
            instance.com_port = kwargs.get("port", PORT)
        instance.is_open = True
        instance.open.side_effect = lambda: setattr(instance, "is_open", True)
        instance.close.side_effect = lambda: setattr(instance, "is_open", False)
        serial_instances.append(instance)
        return instance

    mock_serial_cls.return_value = primary_instance
    mock_serial_cls.side_effect = _make_serial_instance
    yield primary_instance


@pytest.fixture(autouse=True)
def clear_shared_devices():
    SerialDevice._shared_devices.clear()
    yield
    SerialDevice._shared_devices.clear()


@pytest.fixture
def valve_fixture(mock_serial, mocker):
    with patch.object(JKemNewValve, "connect", autospec=True) as connect_spy:
        valve = JKemNewValve(
            com_port=PORT,
            ports={"A": 1, "B": 2, "C": 3},
            valve_num=1,
        )
        connect_spy.assert_called_once_with(valve, 1)
    return valve


def test_initialization_defaults(mock_serial, valve_fixture):
    assert valve_fixture.com_port == PORT
    assert valve_fixture.valve_num == 1
    assert valve_fixture.num_port == 8
    assert valve_fixture.ports == {"A": 1, "B": 2, "C": 3}
    assert valve_fixture.device._connection is mock_serial

    mock_serial._factory.assert_called_once_with(
        port=PORT,
        baudrate=38400,
        parity=valve_fixture.parity,
        bytesize=valve_fixture.bytesize,
        timeout=1.0,
        stopbits=valve_fixture.stopbits,
    )
    mock_serial.close.assert_called_once()


def test_initialization_skips_connect_when_deferred(mock_serial, mocker):
    connect_spy = mocker.patch.object(JKemNewValve, "connect", autospec=True)

    JKemNewValve(com_port=PORT, ports={"A": 1}, valve_num=2, connect_hardware=False)

    connect_spy.assert_not_called()


def test_calculate_checksum_appends_little_endian_sum():
    raw_cmd = bytearray([204, 1, 74, 0, 0, 221])

    result = JKemNewValve.calculate_checksum(raw_cmd)

    assert result == raw_cmd + pack("<H", sum(raw_cmd))


def test_connect_sends_expected_command(mocker, valve_fixture):
    query_spy = mocker.patch.object(valve_fixture, "query", return_value=b"\x00")

    result = valve_fixture.connect(valve_num=1)

    assert result == 0
    query_spy.assert_called_once_with(
        write_command=JKemNewValve.calculate_checksum(bytearray([204, 1, 74, 0, 0, 221])),
        read_until="",
        remove_from_start=2,
        remove_from_end=5,
    )


def test_connect_raises_for_invalid_response_length(mocker, valve_fixture):
    mocker.patch.object(valve_fixture, "query", return_value=b"")

    with pytest.raises(IOError, match="Connecting valve failed!"):
        valve_fixture.connect(valve_num=1)


def test_execute_valve_returns_zero_on_success(mocker, valve_fixture):
    query_spy = mocker.patch.object(valve_fixture, "query", return_value=b"12345678")

    result = valve_fixture._execute_valve(valve_num=1, port=2)

    assert result == 0
    query_spy.assert_called_once_with(
        write_command=JKemNewValve.calculate_checksum(bytearray([204, 1, 68, 2, 0, 221])),
        read_until="",
    )


def test_execute_valve_returns_minus_one_on_short_response(mocker, valve_fixture):
    mocker.patch.object(valve_fixture, "query", return_value=b"1234")

    assert valve_fixture._execute_valve(valve_num=1, port=2) == -1


def test_query_valve_returns_port_number(mocker, valve_fixture):
    query_spy = mocker.patch.object(valve_fixture, "query", return_value=b"\x03")

    result = valve_fixture._query_valve(valve_num=1)

    assert result == 3
    query_spy.assert_called_once_with(
        write_command=JKemNewValve.calculate_checksum(bytearray([204, 1, 62, 0, 0, 221])),
        read_until="",
        remove_from_start=3,
        remove_from_end=4,
    )


def test_query_valve_returns_minus_one_for_invalid_payload_length(mocker, valve_fixture):
    mocker.patch.object(valve_fixture, "query", return_value=b"")

    assert valve_fixture._query_valve(valve_num=1) == -1


def test_get_current_port_returns_query_result(mocker, valve_fixture):
    query_spy = mocker.patch.object(valve_fixture, "_query_valve", return_value=2)

    assert valve_fixture.get_current_port(valve_num=1) == 2

    query_spy.assert_called_once_with(valve_num=1)


def test_get_current_port_retries_then_raises(mocker, valve_fixture):
    query_spy = mocker.patch.object(valve_fixture, "_query_valve", side_effect=ValueError)
    sleep_spy = mocker.patch("matterlab_valves.jkem_new.time.sleep")

    with pytest.raises(IOError, match=f"Query port of valve {PORT} No. 1 exceed max try."):
        valve_fixture.get_current_port(valve_num=1, max_try=3)

    assert query_spy.call_count == 3
    assert sleep_spy.call_count == 3


def test_set_current_port_executes_and_logs_success(mocker, valve_fixture):
    execute_spy = mocker.patch.object(valve_fixture, "_execute_valve")
    get_port_spy = mocker.patch.object(valve_fixture, "get_current_port", return_value=2)
    info_spy = mocker.patch.object(valve_fixture.logger, "info")
    sleep_spy = mocker.patch("matterlab_valves.jkem_new.time.sleep")

    result = valve_fixture.set_current_port(valve_num=1, port=2)

    assert result is None
    execute_spy.assert_called_once_with(valve_num=1, port=2)
    get_port_spy.assert_called_once_with(valve_num=1)
    info_spy.assert_any_call("Requesting JKem valve move on %s No. %s to port %s", PORT, 1, 2)
    info_spy.assert_any_call("Valve on %s No. %s has been moved to %s", PORT, 1, 2)
    sleep_spy.assert_called_once_with(1)


def test_set_current_port_retries_then_raises(mocker, valve_fixture):
    execute_spy = mocker.patch.object(valve_fixture, "_execute_valve")
    get_port_spy = mocker.patch.object(valve_fixture, "get_current_port", return_value=3)
    sleep_spy = mocker.patch("matterlab_valves.jkem_new.time.sleep")

    with pytest.raises(IOError, match=f"Move port of valve {PORT} No. 1 exceed max try."):
        valve_fixture.set_current_port(valve_num=1, port=2, max_try=3)

    assert execute_spy.call_count == 3
    assert get_port_spy.call_count == 3
    assert sleep_spy.call_count == 3


def test_port_property_reads_from_active_valve(mocker, valve_fixture):
    get_port_spy = mocker.patch.object(valve_fixture, "get_current_port", return_value=2)

    assert valve_fixture.port == 2

    get_port_spy.assert_called_once_with(valve_num=1)


def test_port_property_sets_active_valve_port(mocker, valve_fixture):
    set_port_spy = mocker.patch.object(valve_fixture, "set_current_port")

    valve_fixture.port = 3

    set_port_spy.assert_called_once_with(valve_num=1, port=3)



from pathlib import Path
import time
from logging import Logger
from typing import Dict, Tuple, Optional

from matterlab_serial_device import SerialDevice, open_close

from matterlab_valves.base_valve import SwitchingValve


class ValcoSelectionValve(SwitchingValve, SerialDevice):
    category="Valve"
    ui_fields  = ("com_port", "num_port")
    def __init__(self,
                 com_port: str,
                 num_port: int,
                 encoding: str = "utf-8",
                 baudrate: int = 9600,
                 timeout: float = 1.0,
                 parity: str = "none",
                 bytesize: int = 8,
                 stopbits: int = 1,
                 direction_CW: bool = True,
                 ports: Optional[Dict[str, int]] = None,
                 connect_hardware: bool = True,
                 logger: Optional[Logger] = None,
                 **kwargs
                 ):
        """
        :param com_port: COM port of the pump connected to, example: "COM1", "/tty/USB0"
        :param ports: dict containing nickname and valve port number mapping
        :param connect_hardware: if connect hardware up on initialization                                         
        :return:
        """

        SerialDevice.__init__(self,
                              com_port=com_port,
                              encoding=encoding,
                              baudrate=baudrate,
                              timeout=timeout,
                              parity=parity,
                              bytesize=bytesize,
                              stopbits=stopbits,
                              **kwargs
                              )
        SwitchingValve.__init__(self, 
                                num_port=num_port,
                                ports=ports,
                                logger=logger
                                )
        self.direction_CW = direction_CW
        if connect_hardware:
            self.connect()

    def connect(self):
        """
        connect the device, hardware initialization method
        :return:
        """
        self.logger.info("Connecting Valco valve on %s", self.com_port)
        self._set_response_mode()
        self._set_device_mode()
        self.num_port = (self.num_port, False)
        self._home_valve()
        self.logger.info("Connected Valco valve on %s with %s ports", self.com_port, self.num_port)

    @open_close
    def _query_valve(self, command: str) -> str:
        return self.query(write_command=f'{command}\r',
                          read_until=b'\r',
                          remove_from_start=2,
                          remove_from_end=1)

    @open_close
    def _execute_valve(self, command: str):
        """
        execute operations on valve
        """
        self.write(command=f'{command}\r')

    def _set_response_mode(self):
        """
        set the response mode to basic string
        """
        self._execute_valve('IFM1')

    def _query_device_mode(self):
        return self._query_valve(command='AM')

    def _set_device_mode(self, max_trial=5):
        """
        set the device mode to selection valve
        maximum try max_trial times
        """
        for i in range(0, max_trial):
            self._execute_valve('AM3')
            rtn = self._query_device_mode()
            if rtn == '3':
                return
            else:
                continue
        raise IOError('Failed to set device mode to selection valve')

    @property
    def num_port(self) -> int:
        return int(self._query_valve(command='NP'), base=10)

    @num_port.setter
    def num_port(self, new_num_port: int, reinstall: bool = False):
        """
        set the number of port based on physical port number
        essential if change the hardware configuration
        """
        if reinstall:
            self._execute_valve('AL')
            input(f"Resetting the number of port, press Enter to confirm {new_num_port} port valve is installed.")
        self._execute_valve(f'NP{new_num_port}')

    def get_current_port(self, max_try: int = 5) -> int:
        """
        report current port position as int
        Returns
        -------

        """
        for i in range(0, max_try):
            try:
                port = int(self._query_valve(command='CP'), 10)
                self.logger.debug("Valco valve on %s reports current port %s", self.com_port, port)
                return port
            except:
                time.sleep(1)
                continue
        raise IOError(f'Query port of valve on {self.com_port} exceed max try.')

    def set_current_port(self, port: int, direction_CW: bool = None, max_try=5):
        """
        move the valve port to position
        move direction default in clock wise
        try to move trial times, if all failed raise error
        """
        assert isinstance(port, int), "port number must be int!"
        assert 0 < port <= self.num_port, "Valve port number out of range!"
        self.logger.info(
            "Requesting Valco valve move on %s to port %s (direction_CW=%s)",
            self.com_port,
            port,
            direction_CW,
        )
        # if direction_CW is None:
        #     direction_CW = self.direction_CW
        for i in range(0, max_try):
            if direction_CW is None:
                self._execute_valve(f'GO{port}')
            else:
                if direction_CW:
                    self._execute_valve(f'CW{port}')
                else:
                    self._execute_valve(f'CC{port}')
            if self.get_current_port() == port:
                self.logger.info(
                    "Valve on %s has been moved to %s position.",
                    self.com_port,
                    port,
                )
                return None
            else:
                time.sleep(1)
                continue
        raise IOError(f'Move port of valve on {self.com_port} exceed max try.')

    def _home_valve(self):
        """
        home valve
        """
        self.logger.info("Homing Valco valve on %s", self.com_port)
        self._execute_valve('HM')

    @property
    def port(self) -> int:
        """
        concrete method to report the port of valve
        """
        return self.get_current_port()

    @port.setter
    def port(self, port: int):
        """
        concrete method to set the port of valve
        """
        self.set_current_port(port = port)



import os

import pytest

from matterlab_serial_device import SerialDevice
from matterlab_valves.runze_valve import RunzeMotorEventType, RunzeSelectionValve, RunzeValveError


PORT = "/dev/ttyUSB0" if os.name == "posix" else "COM1"


@pytest.fixture
def mock_serial(mocker):
    mock_serial_cls = mocker.patch("matterlab_serial_device.transport.serial.Serial", autospec=True)
    serial_instances = []
    primary_instance = mocker.MagicMock()
    primary_instance.com_port = PORT
    primary_instance._factory = mock_serial_cls
    primary_instance._instances = serial_instances

    def _make_serial_instance(*args, **kwargs):
        if not serial_instances:
            instance = primary_instance
        else:
            instance = mocker.MagicMock()
            instance.com_port = kwargs.get("port", PORT)
        instance.is_open = True
        instance.open.side_effect = lambda: setattr(instance, "is_open", True)
        instance.close.side_effect = lambda: setattr(instance, "is_open", False)
        serial_instances.append(instance)
        return instance

    mock_serial_cls.return_value = primary_instance
    mock_serial_cls.side_effect = _make_serial_instance
    yield primary_instance


@pytest.fixture(autouse=True)
def clear_shared_devices():
    SerialDevice._shared_devices.clear()
    yield
    SerialDevice._shared_devices.clear()


@pytest.fixture
def valve_fixture(mock_serial):
    return RunzeSelectionValve(
        com_port=PORT,
        address=0,
        num_port=10,
        ports={"home": 1, "collect": 2},
    )


def test_initialization_defaults(mock_serial, valve_fixture):
    assert valve_fixture.com_port == PORT
    assert valve_fixture.address == 0
    assert valve_fixture.num_port == 10
    assert valve_fixture.direction_CW is True
    assert valve_fixture.share_port is True
    assert valve_fixture.ports == {"home": 1, "collect": 2}
    assert valve_fixture.device._connection is mock_serial

    mock_serial._factory.assert_called_once_with(
        port=PORT,
        baudrate=9600,
        parity=valve_fixture.parity,
        bytesize=valve_fixture.bytesize,
        timeout=1.0,
        stopbits=valve_fixture.stopbits,
    )


def test_calc_checksum_returns_little_endian_sum():
    raw_cmd = bytearray([0xCC, 0x00, 0x3E, 0x00, 0x00, 0xDD])

    assert RunzeSelectionValve.calc_checksum(raw_cmd) == bytearray([0xE7, 0x01])


def test_execute_valve_wraps_command_and_returns_bytes(mocker, valve_fixture):
    query_spy = mocker.patch.object(valve_fixture, "query", return_value=b"\xcc\x00\xfe\x02\xdd")

    result = valve_fixture._execute_valve(bytearray([0x00, 0x44, 0x02, 0x00]))

    expected_command = bytearray([0xCC, 0x00, 0x44, 0x02, 0x00, 0xDD])
    expected_command += RunzeSelectionValve.calc_checksum(expected_command)
    assert result == bytearray(b"\xcc\x00\xfe\x02\xdd")
    query_spy.assert_called_once_with(write_command=expected_command, return_bytes=True)


def test_execute_valve_raises_on_empty_response(mocker, valve_fixture):
    mocker.patch.object(valve_fixture, "query", return_value=b"")

    with pytest.raises(RunzeValveError, match="Return 0 length data, communication failed"):
        valve_fixture._execute_valve(bytearray([0x00, 0x44, 0x02, 0x00]))


def test_set_address_sends_command_after_confirmation(mocker, valve_fixture):
    execute_spy = mocker.patch.object(valve_fixture, "_execute_valve")
    warning_spy = mocker.patch.object(valve_fixture.logger, "warning")
    mocker.patch("builtins.input", return_value="Yes")

    valve_fixture.set_address(5)

    execute_spy.assert_called_once_with(bytearray([0, 0x00, 0xFF, 0xEE, 0xBB, 0xAA, 5, 0x00, 0x00, 0x00]))
    warning_spy.assert_called_once()


def test_set_address_returns_early_when_not_confirmed(mocker, valve_fixture):
    execute_spy = mocker.patch.object(valve_fixture, "_execute_valve")
    warning_spy = mocker.patch.object(valve_fixture.logger, "warning")
    mocker.patch("builtins.input", return_value="No")

    valve_fixture.set_address(5)

    execute_spy.assert_not_called()
    warning_spy.assert_called_once_with("Set RS485 address cancelled.")


def test_set_address_raises_for_out_of_range_value(mocker, valve_fixture):
    mocker.patch("builtins.input", return_value="Yes")

    with pytest.raises(ValueError, match="RS485 address out of range \\[0, 0x7f\\]"):
        valve_fixture.set_address(0x80)


def test_check_motor_returns_status_value(mocker, valve_fixture):
    execute_spy = mocker.patch.object(valve_fixture, "_execute_valve", return_value=bytearray([0xCC, 0x00, 0x00, 0x00]))

    assert valve_fixture.check_motor() == RunzeMotorEventType.Normal_status.value

    execute_spy.assert_called_once_with(bytearray([0, 0x4A, 0x00, 0x00]))


def test_check_motor_raises_for_unknown_status(mocker, valve_fixture):
    mocker.patch.object(valve_fixture, "_execute_valve", return_value=bytearray([0xCC, 0x00, 0x99, 0x00]))

    with pytest.raises(RunzeValveError, match="Unknown motor status"):
        valve_fixture.check_motor()


def test_get_current_port_queries_device(mocker, valve_fixture):
    execute_spy = mocker.patch.object(valve_fixture, "_execute_valve", return_value=bytearray([0xCC, 0x00, 0x00, 3]))

    assert valve_fixture.get_current_port() == 3
    assert valve_fixture._port == 3

    execute_spy.assert_called_once_with(bytearray([0, 0x3E, 0x00, 0x00]))


def test_set_current_port_uses_direct_move_by_default(mocker, valve_fixture):
    execute_spy = mocker.patch.object(valve_fixture, "_execute_valve", return_value=bytearray([0xCC, 0x00, 0xFE, 0x00]))
    get_port_spy = mocker.patch.object(valve_fixture, "get_current_port", return_value=2)
    sleep_spy = mocker.patch("matterlab_valves.runze_valve.time.sleep")

    valve_fixture.set_current_port(port=2)

    execute_spy.assert_called_once_with(bytearray([0, 0x44, 2, 0x00]))
    assert get_port_spy.call_count == 2
    sleep_spy.assert_called_once_with(0.5)


def test_set_current_port_uses_clockwise_through_port(mocker, valve_fixture):
    execute_spy = mocker.patch.object(valve_fixture, "_execute_valve", return_value=bytearray([0xCC, 0x00, 0xFE, 0x00]))
    get_port_spy = mocker.patch.object(valve_fixture, "get_current_port", return_value=2)
    mocker.patch("matterlab_valves.runze_valve.time.sleep")

    valve_fixture.set_current_port(port=2, direction_CW=True)

    execute_spy.assert_called_once_with(bytearray([0, 0xA4, 3, 2]))
    assert get_port_spy.call_count == 2


def test_set_current_port_uses_counter_clockwise_through_port(mocker, valve_fixture):
    execute_spy = mocker.patch.object(valve_fixture, "_execute_valve", return_value=bytearray([0xCC, 0x00, 0xFE, 0x00]))
    get_port_spy = mocker.patch.object(valve_fixture, "get_current_port", return_value=2)
    mocker.patch("matterlab_valves.runze_valve.time.sleep")

    valve_fixture.set_current_port(port=2, direction_CW=False)

    execute_spy.assert_called_once_with(bytearray([0, 0xA4, 1, 2]))
    assert get_port_spy.call_count == 2


def test_set_current_port_raises_for_out_of_range_port(valve_fixture):
    with pytest.raises(ValueError, match="Port number out of range, default \\[1, 10\\], set to 11"):
        valve_fixture.set_current_port(port=11)


def test_set_current_port_raises_when_task_not_accepted(mocker, valve_fixture):
    mocker.patch.object(valve_fixture, "_execute_valve", return_value=bytearray([0xCC, 0x00, 0x00, 0x00]))

    with pytest.raises(RunzeValveError, match="Set port failed, did not return 'Task being executed'"):
        valve_fixture.set_current_port(port=2, confirmation=False)


def test_set_current_port_raises_when_confirmation_fails(mocker, valve_fixture):
    mocker.patch.object(valve_fixture, "_execute_valve", return_value=bytearray([0xCC, 0x00, 0xFE, 0x00]))
    mocker.patch.object(valve_fixture, "get_current_port", return_value=3)
    mocker.patch("matterlab_valves.runze_valve.time.sleep")

    with pytest.raises(RunzeValveError, match="Set port failed: Expected 2, got 3"):
        valve_fixture.set_current_port(port=2)


def test_set_current_port_updates_cached_port_without_confirmation(mocker, valve_fixture):
    execute_spy = mocker.patch.object(valve_fixture, "_execute_valve", return_value=bytearray([0xCC, 0x00, 0xFE, 0x00]))
    get_port_spy = mocker.patch.object(valve_fixture, "get_current_port")

    valve_fixture.set_current_port(port=4, confirmation=False)

    execute_spy.assert_called_once_with(bytearray([0, 0x44, 4, 0x00]))
    get_port_spy.assert_called_once_with()
    assert valve_fixture._port == 4


def test_reset_moves_home(mocker, valve_fixture):
    execute_spy = mocker.patch.object(valve_fixture, "_execute_valve")

    valve_fixture.reset(origin=False)

    execute_spy.assert_called_once_with(bytearray([0, 0x45, 0x00, 0x00]))
    assert valve_fixture._port == 1


def test_reset_origin_uses_origin_command(mocker, valve_fixture):
    execute_spy = mocker.patch.object(valve_fixture, "_execute_valve")

    valve_fixture.reset(origin=True)

    execute_spy.assert_called_once_with(bytearray([0, 0x4F, 0x00, 0x00]))
    assert valve_fixture._port == 1


def test_set_between_ports_uses_adjacent_port(mocker, valve_fixture):
    execute_spy = mocker.patch.object(valve_fixture, "_execute_valve")

    valve_fixture.set_between_ports(3)

    execute_spy.assert_called_once_with(bytearray([0, 0xB4, 3, 4]))


def test_set_between_ports_wraps_from_last_port(mocker, valve_fixture):
    execute_spy = mocker.patch.object(valve_fixture, "_execute_valve")

    valve_fixture.set_between_ports(10)

    execute_spy.assert_called_once_with(bytearray([0, 0xB4, 10, 1]))


def test_set_between_ports_raises_for_out_of_range_port(valve_fixture):
    with pytest.raises(ValueError, match="Port number out of range, default \\[1, 10\\], set to 11"):
        valve_fixture.set_between_ports(11)


def test_home_valve_calls_reset(mocker, valve_fixture):
    reset_spy = mocker.patch.object(valve_fixture, "reset")

    valve_fixture._home_valve()

    reset_spy.assert_called_once_with(origin=False)


def test_port_property_reads_current_port(mocker, valve_fixture):
    get_port_spy = mocker.patch.object(valve_fixture, "get_current_port", return_value=2)

    assert valve_fixture.port == 2

    get_port_spy.assert_called_once_with()


def test_port_property_sets_current_port(mocker, valve_fixture):
    set_port_spy = mocker.patch.object(valve_fixture, "set_current_port")

    valve_fixture.port = 3

    set_port_spy.assert_called_once_with(port=3)



from __future__ import annotations

from dataclasses import dataclass

from matterlab_serial_device import ScriptedSerialTransport
from matterlab_serial_device.transport import SerialTransportConfig

from matterlab_valves.runze_valve import RunzeMotorEventType, RunzeSelectionValve


@dataclass
class RunzeSelectionSimState:
    address: int
    num_port: int
    current_port: int = 1
    motor_status: RunzeMotorEventType = RunzeMotorEventType.Normal_status
    between_ports_after: int | None = None


class RunzeSelectionSimProtocol:
    def __init__(self, state: RunzeSelectionSimState) -> None:
        self.state = state

    def __call__(self, write_data: bytes, expected: bytes, size: int | None) -> bytes:
        del expected, size
        if len(write_data) < 7:
            return b""
        if write_data[0] != RunzeSelectionValve.STX or write_data[-3] != RunzeSelectionValve.ETX:
            return b""

        body = write_data[1:-3]
        address = body[0]
        command = body[1]
        data_1 = body[2] if len(body) > 2 else 0
        data_2 = body[3] if len(body) > 3 else 0

        if address != self.state.address:
            return b""

        if command == 0x3E:
            return bytes([RunzeSelectionValve.STX, address, RunzeMotorEventType.Normal_status.value, self.state.current_port])

        if command == 0x4A:
            return bytes([RunzeSelectionValve.STX, address, self.state.motor_status.value, self.state.current_port])

        if command == 0x44:
            self._set_port(data_1)
            return bytes([RunzeSelectionValve.STX, address, RunzeMotorEventType.Task_being_executed.value, self.state.current_port])

        if command == 0xA4:
            self._set_port(data_2)
            return bytes([RunzeSelectionValve.STX, address, RunzeMotorEventType.Task_being_executed.value, self.state.current_port])

        if command in {0x45, 0x4F}:
            self.state.current_port = 1
            self.state.between_ports_after = None
            return bytes([RunzeSelectionValve.STX, address, RunzeMotorEventType.Task_being_executed.value, self.state.current_port])

        if command == 0xB4:
            self._validate_port(data_1)
            self.state.between_ports_after = data_1
            return bytes([RunzeSelectionValve.STX, address, RunzeMotorEventType.Task_being_executed.value, data_1])

        if command == 0x00 and len(body) >= 7:
            self.state.address = body[6]
            return bytes([RunzeSelectionValve.STX, address, RunzeMotorEventType.Task_being_executed.value, self.state.current_port])

        if command == 0x02 and len(body) >= 7:
            return bytes([RunzeSelectionValve.STX, address, RunzeMotorEventType.Task_being_executed.value, self.state.current_port])

        if command == 0xFF:
            self.state.current_port = 1
            self.state.motor_status = RunzeMotorEventType.Normal_status
            self.state.between_ports_after = None
            return bytes([RunzeSelectionValve.STX, address, RunzeMotorEventType.Task_being_executed.value, self.state.current_port])

        return bytes([RunzeSelectionValve.STX, address, RunzeMotorEventType.Parameter_error.value, self.state.current_port])

    def _validate_port(self, port: int) -> None:
        if not 1 <= port <= self.state.num_port:
            raise ValueError(f"Port number out of range, default [1, {self.state.num_port}], set to {port}")

    def _set_port(self, port: int) -> None:
        self._validate_port(port)
        self.state.current_port = port
        self.state.between_ports_after = None


def create_runze_selection_sim_transport_factory(
    address: int,
    num_port: int,
    *,
    initial_port: int = 1,
    motor_status: RunzeMotorEventType = RunzeMotorEventType.Normal_status,
):
    state = RunzeSelectionSimState(
        address=address,
        num_port=num_port,
        current_port=initial_port,
        motor_status=motor_status,
    )
    protocol = RunzeSelectionSimProtocol(state)

    def _factory(config: SerialTransportConfig):
        return ScriptedSerialTransport(config, default_response=protocol)

    _factory.state = state
    return _factory



from abc import ABC, abstractmethod
from logging import Logger, getLogger
from typing import Dict, Optional, Union


class SwitchingValve(ABC):
    # TODO: Should SwitchingValve have a logger?
    """
    Abstract Base Class for handling different kinds of switching valves.
    """
    category="Valve"
    ui_fields  = ("com_port", "num_port")
    def __init__(self,
                 num_port: int,
                 ports: Optional[Dict[str, int]] = None,
                 logger: Optional[Logger] = None) -> None:
        """
        Args:
            ports: Dictionary of human identifiers for the ports, with the respective port numbers as values.
        """
        self.num_port = num_port
        self.ports: Optional[Dict[str, int]] = ports if ports else {}
        self.logger = logger if logger is not None else getLogger(
            f"{self.__class__.__module__}.{self.__class__.__name__}"
        )

    # Public Methods for Using the Switching Valve
    @abstractmethod
    def port(self) -> int:
        """
        Abstract method that reads the current valve position.
        """
        pass

    def switch_port(self, port: Union[str, int]) -> int:
        """
        If applicable, converts the human identifier for a port (as key in the ports attribute) to the respective
        computer identifier.

        Args:
            port: Human identifier for the port.

        Returns:
            int: Port number, as required for communication with the pump.
        """
        if isinstance(port, str):
            port = self.ports.get(port)
        if port is not None:
            self.port = port
            return port
        else:
            raise ValueError('Wrong ports')

    # Private methods for translating the port id and name
    def _get_port_id(self, port_num: int) -> Union[str, int]:
        """
        Retrieve the port ID based on port number.

        Args:
            port_num: Port number.

        Returns:
            port_id: Port ID.
        """
        for port_id, port in self.ports.items():
            if port == port_num:
                return port_id
        raise KeyError(f'Port number {port_num} not found in ports.')




from .runze_selection_sim import (
    RunzeSelectionSimProtocol,
    RunzeSelectionSimState,
    create_runze_selection_sim_transport_factory,
)

__all__ = [
    "RunzeSelectionSimProtocol",
    "RunzeSelectionSimState",
    "create_runze_selection_sim_transport_factory",
]



from .valco_selection_valve import ValcoSelectionValve
from .jkem_4in1_valve import JKem4in1Valve
from .jkem_new import JKemNewValve
from .runze_valve import RunzeSelectionValve
from .base_valve import SwitchingValve
from .sim.runze_selection_sim import create_runze_selection_sim_transport_factory

__all__ = [
    "ValcoSelectionValve",
    "JKem4in1Valve",
    "JKemNewValve",
    "SwitchingValve",
    "RunzeSelectionValve",
    "create_runze_selection_sim_transport_factory",
]



import pytest

from matterlab_valves import ValcoSelectionValve


@pytest.mark.real
def test_valco_selection_real_read(hardware_com_port, hardware_num_port):
    valve = ValcoSelectionValve(
        com_port=hardware_com_port,
        num_port=hardware_num_port,
        ports={"home": 1},
    )

    assert valve.num_port == hardware_num_port
    assert isinstance(valve.port, int)


@pytest.mark.real
def test_valco_selection_real_switch(hardware_com_port, hardware_num_port):
    valve = ValcoSelectionValve(
        com_port=hardware_com_port,
        num_port=hardware_num_port,
        ports={"home": 1},
    )

    valve.port = 2
    assert valve.port == 2

    valve.switch_port("home")
    assert valve.port == 1



import os

import pytest

from matterlab_serial_device import SerialDevice
from matterlab_valves.valco_selection_valve import ValcoSelectionValve


PORT = "/dev/ttyUSB0" if os.name == "posix" else "COM1"


@pytest.fixture
def mock_serial(mocker):
    mock_serial_cls = mocker.patch("matterlab_serial_device.transport.serial.Serial", autospec=True)
    serial_instances = []
    primary_instance = mocker.MagicMock()
    primary_instance.com_port = PORT
    primary_instance._factory = mock_serial_cls
    primary_instance._instances = serial_instances

    def _make_serial_instance(*args, **kwargs):
        if not serial_instances:
            instance = primary_instance
        else:
            instance = mocker.MagicMock()
            instance.com_port = kwargs.get("port", PORT)
        instance.is_open = True
        instance.open.side_effect = lambda: setattr(instance, "is_open", True)
        instance.close.side_effect = lambda: setattr(instance, "is_open", False)
        serial_instances.append(instance)
        return instance

    mock_serial_cls.return_value = primary_instance
    mock_serial_cls.side_effect = _make_serial_instance
    yield primary_instance


@pytest.fixture(autouse=True)
def clear_shared_devices():
    SerialDevice._shared_devices.clear()
    yield
    SerialDevice._shared_devices.clear()


@pytest.fixture
def valve_fixture(mock_serial, mocker):
    query_spy = mocker.patch("matterlab_serial_device.serial_device.SerialDevice.query")
    query_spy.return_value = "3"
    return ValcoSelectionValve(
        com_port=PORT,
        num_port=3,
        ports={"A": 1, "B": 2, "C": 3},
    )


def test_initialization_defaults(mock_serial, valve_fixture):
    assert valve_fixture.com_port == PORT
    assert valve_fixture.direction_CW is True
    assert valve_fixture.ports == {"A": 1, "B": 2, "C": 3}
    assert valve_fixture.device._connection is mock_serial

    mock_serial._factory.assert_called_once_with(
        port=PORT,
        baudrate=9600,
        parity=valve_fixture.parity,
        bytesize=valve_fixture.bytesize,
        timeout=1.0,
        stopbits=valve_fixture.stopbits,
    )
    assert mock_serial.close.call_count >= 1


def test_initialization_calls_connect_when_requested(mock_serial, mocker):
    connect_spy = mocker.patch.object(ValcoSelectionValve, "connect", autospec=True)

    ValcoSelectionValve(com_port=PORT, num_port=3, ports={"A": 1}, connect_hardware=True)

    connect_spy.assert_called_once()


def test_initialization_skips_connect_when_deferred(mock_serial, mocker):
    connect_spy = mocker.patch.object(ValcoSelectionValve, "connect", autospec=True)

    ValcoSelectionValve(com_port=PORT, num_port=3, ports={"A": 1}, connect_hardware=False)

    connect_spy.assert_not_called()


def test_connect_initializes_valve(mocker, valve_fixture):
    set_response_mode_spy = mocker.patch.object(valve_fixture, "_set_response_mode")
    set_device_mode_spy = mocker.patch.object(valve_fixture, "_set_device_mode")
    home_spy = mocker.patch.object(valve_fixture, "_home_valve")
    num_port_spy = mocker.patch.object(type(valve_fixture), "num_port", new_callable=mocker.PropertyMock, return_value=3)

    valve_fixture.connect()

    set_response_mode_spy.assert_called_once_with()
    set_device_mode_spy.assert_called_once_with()
    home_spy.assert_called_once_with()
    assert num_port_spy.call_count >= 2


def test_query_valve_uses_expected_wire_format(mocker, valve_fixture):
    query_spy = mocker.patch.object(valve_fixture, "query", return_value="response")

    result = valve_fixture._query_valve("CP")

    assert result == "response"
    query_spy.assert_called_once_with(
        write_command="CP\r",
        read_until=b"\r",
        remove_from_start=2,
        remove_from_end=1,
    )


def test_execute_valve_writes_expected_command(mocker, valve_fixture):
    write_spy = mocker.patch.object(valve_fixture, "write")

    valve_fixture._execute_valve("HM")

    write_spy.assert_called_once_with(command="HM\r")


def test_set_response_mode_executes_ifm1(mocker, valve_fixture):
    execute_spy = mocker.patch.object(valve_fixture, "_execute_valve")

    valve_fixture._set_response_mode()

    execute_spy.assert_called_once_with("IFM1")


def test_query_device_mode_uses_am_command(mocker, valve_fixture):
    query_spy = mocker.patch.object(valve_fixture, "_query_valve", return_value="3")

    assert valve_fixture._query_device_mode() == "3"

    query_spy.assert_called_once_with(command="AM")


def test_set_device_mode_retries_until_success(mocker, valve_fixture):
    execute_spy = mocker.patch.object(valve_fixture, "_execute_valve")
    query_device_mode_spy = mocker.patch.object(valve_fixture, "_query_device_mode", side_effect=["2", "3"])

    valve_fixture._set_device_mode(max_trial=3)

    assert execute_spy.call_count == 2
    assert query_device_mode_spy.call_count == 2


def test_set_device_mode_raises_after_max_trials(mocker, valve_fixture):
    mocker.patch.object(valve_fixture, "_execute_valve")
    mocker.patch.object(valve_fixture, "_query_device_mode", return_value="2")

    with pytest.raises(IOError, match="Failed to set device mode to selection valve"):
        valve_fixture._set_device_mode(max_trial=3)


def test_num_port_getter_queries_np(mocker, valve_fixture):
    query_spy = mocker.patch.object(valve_fixture, "_query_valve", return_value="6")

    assert valve_fixture.num_port == 6

    query_spy.assert_called_once_with(command="NP")


def test_num_port_setter_sends_np_command(mocker, valve_fixture):
    execute_spy = mocker.patch.object(valve_fixture, "_execute_valve")

    type(valve_fixture).num_port.fset(valve_fixture, 10)

    execute_spy.assert_called_once_with("NP10")


def test_home_valve_executes_hm(mocker, valve_fixture):
    execute_spy = mocker.patch.object(valve_fixture, "_execute_valve")

    valve_fixture._home_valve()

    execute_spy.assert_called_once_with("HM")


@pytest.mark.parametrize(
    ("port", "direction_cw", "expected_command"),
    [
        (2, None, "GO2"),
        (2, True, "CW2"),
        (3, False, "CC3"),
    ],
)
def test_set_current_port_executes_expected_command(mocker, valve_fixture, port, direction_cw, expected_command):
    execute_spy = mocker.patch.object(valve_fixture, "_execute_valve")
    get_port_spy = mocker.patch.object(valve_fixture, "get_current_port", return_value=port)
    info_spy = mocker.patch.object(valve_fixture.logger, "info")
    mocker.patch.object(type(valve_fixture), "num_port", new_callable=mocker.PropertyMock, return_value=10)

    result = valve_fixture.set_current_port(port=port, direction_CW=direction_cw)

    assert result is None
    execute_spy.assert_called_once_with(expected_command)
    get_port_spy.assert_called_once_with()
    info_spy.assert_any_call(
        "Requesting Valco valve move on %s to port %s (direction_CW=%s)",
        PORT,
        port,
        direction_cw,
    )
    info_spy.assert_any_call("Valve on %s has been moved to %s position.", PORT, port)


def test_set_current_port_raises_for_non_integer(valve_fixture):
    with pytest.raises(AssertionError, match="port number must be int!"):
        valve_fixture.set_current_port("invalid")


def test_set_current_port_raises_for_out_of_range_port(mocker, valve_fixture):
    mocker.patch.object(type(valve_fixture), "num_port", new_callable=mocker.PropertyMock, return_value=3)

    with pytest.raises(AssertionError, match="Valve port number out of range!"):
        valve_fixture.set_current_port(10)


def test_set_current_port_retries_then_raises(mocker, valve_fixture):
    execute_spy = mocker.patch.object(valve_fixture, "_execute_valve")
    get_port_spy = mocker.patch.object(valve_fixture, "get_current_port", return_value=3)
    sleep_spy = mocker.patch("matterlab_valves.valco_selection_valve.time.sleep")
    mocker.patch.object(type(valve_fixture), "num_port", new_callable=mocker.PropertyMock, return_value=10)

    with pytest.raises(IOError, match=f"Move port of valve on {PORT} exceed max try."):
        valve_fixture.set_current_port(port=2, max_try=3)

    assert execute_spy.call_count == 3
    assert get_port_spy.call_count == 3
    assert sleep_spy.call_count == 3


def test_get_current_port_returns_integer(mocker, valve_fixture):
    query_spy = mocker.patch.object(valve_fixture, "_query_valve", return_value="2")

    assert valve_fixture.get_current_port() == 2

    query_spy.assert_called_once_with(command="CP")


def test_get_current_port_retries_then_raises(mocker, valve_fixture):
    query_spy = mocker.patch.object(valve_fixture, "_query_valve", side_effect=ValueError)
    sleep_spy = mocker.patch("matterlab_valves.valco_selection_valve.time.sleep")

    with pytest.raises(IOError, match=f"Query port of valve on {PORT} exceed max try."):
        valve_fixture.get_current_port(max_try=3)

    assert query_spy.call_count == 3
    assert sleep_spy.call_count == 3


def test_port_property_reads_from_device(mocker, valve_fixture):
    get_port_spy = mocker.patch.object(valve_fixture, "get_current_port", return_value=2)

    assert valve_fixture.port == 2

    get_port_spy.assert_called_once_with()


def test_port_property_sets_active_port(mocker, valve_fixture):
    set_port_spy = mocker.patch.object(valve_fixture, "set_current_port")

    valve_fixture.port = 3

    set_port_spy.assert_called_once_with(port=3)



import pytest

from matterlab_valves import RunzeSelectionValve


@pytest.mark.real
def test_runze_selection_real_read(hardware_com_port, hardware_num_port, hardware_address):
    valve = RunzeSelectionValve(
        com_port=hardware_com_port,
        address=hardware_address,
        num_port=hardware_num_port,
        ports={"home": 1},
    )

    assert valve.num_port == hardware_num_port


@pytest.mark.real
def test_runze_selection_real_switch(hardware_com_port, hardware_num_port, hardware_address):
    valve = RunzeSelectionValve(
        com_port=hardware_com_port,
        address=hardware_address,
        num_port=hardware_num_port,
        ports={"home": 1},
    )

    valve.port = 2
    assert valve.port == 2

    valve.switch_port("home")
    assert valve.port == 1



from pathlib import Path
import time
from logging import Logger
from typing import Dict, Tuple, Optional
from enum import IntEnum
from matterlab_serial_device import SerialDevice, open_close

from matterlab_valves.base_valve import SwitchingValve


class RunzeValveError(Exception):
    pass

class RunzeMotorEventType(IntEnum):
    Normal_status = 0x00
    Frame_error = 0x01
    Parameter_error = 0x02
    Optocoupler_error = 0x03
    Motor_busy = 0x04
    Motor_stalled = 0x05
    Unknown_position = 0x06
    Task_being_executed = 0xfe
    Unknown_error = 0xff

class RunzeSelectionValve(SwitchingValve, SerialDevice):
    category="Valve"
    ui_fields  = ("com_port", "address", "num_port")
    STX = 0xcc
    ETX = 0xdd

    def __init__(self,
                 com_port: str,
                 address: int,
                 num_port: int,
                 encoding: str = "utf-8",
                 baudrate: int = 9600,
                 timeout: float = 1.0,
                 parity: str = "none",
                 bytesize: int = 8,
                 stopbits: int = 1,
                 direction_CW: bool = True,
                 ports: Optional[Dict[str, int]] = None,
                 connect_hardware: bool = True,
                 share_port: bool = True,
                 logger: Optional[Logger] = None,
                 **kwargs
                 ):
        SerialDevice.__init__(
            self,
            com_port=com_port,
            encoding=encoding,
            baudrate=baudrate,
            timeout=timeout,
            parity=parity,
            bytesize=bytesize,
            stopbits=stopbits,
            share_port=share_port,
            **kwargs
        )
        SwitchingValve.__init__(self,
                                num_port=num_port,
                                ports=ports,
                                logger=logger
                                )
        self.address=address
        self.num_port=num_port
        self.direction_CW = direction_CW

    @staticmethod
    def calc_checksum(cmd:bytearray) -> bytearray:
        checksum = 0
        for item in cmd:
            checksum += item
        return bytearray([checksum & 0xff, (checksum >>8) & 0xff])

    @open_close
    def _execute_valve(self, cmd_raw: bytearray) -> bytearray:
        """
        Send execution command to the valve and get response
        :param cmd_raw: raw command
        :return:
        """
        cmd = bytearray([RunzeSelectionValve.STX]) + cmd_raw + bytearray([RunzeSelectionValve.ETX])
        cmd += self.calc_checksum(cmd)
        rtn = self.query(write_command=cmd, return_bytes = True)
        if len(rtn):
            return bytearray(rtn)
        else:
            raise RunzeValveError("Return 0 length data, communication failed")

    def set_address(self, address: int) -> None:
        """
        Set the RS485 address of the valve
        :param address: RS485 address, range [0x00, 0x7F], default 0x00
        :return:
        """
        if "Yes" != input("Changing RS485 address is dangerous, MUST label on the valve right after!\n"
                          "Type Yes for confirmation:\n") :
            self.logger.warning("Set RS485 address cancelled.")
            return
        if not 0x00 <= address <= 0x7f:
            raise ValueError("RS485 address out of range [0, 0x7f]")
        self._execute_valve(bytearray([self.address, 0x00, 0xff, 0xee, 0xbb, 0xaa, address, 0x00,0x00,0x00]))
        self.logger.warning("RS485 address of valve set to %s. Please disconnect power and label the address on the hardware!", address)

    def set_baudrate(self) -> None:
        """
        Set the RS485 baudrate of the valve
        :return:
        """
        if "Yes" != input("Changing RS485 baudrate is dangerous, MUST label on the valve right after!\n"
                          "Type Yes for confirmation:\n"):
            self.logger.warning("Set RS485 baudrate cancelled.")
            return
        baud_options = {"9600": 0x00, "19200": 0x01, "38400": 0x02, "57600": 0x03, "115200": 0x04}
        baud = input("Enter baudrate (9600, 19200, 38400, 57600, 115200):\n")
        if baud not in baud_options:
            raise ValueError("Invalid baudrate entered")
        self._execute_valve(
            bytearray([self.address, 0x02, 0xff, 0xee, 0xbb, 0xaa, baud_options[baud], 0x00, 0x00, 0x00]))
        self.logger.warning("RS485 baudrate of valve set to %s. Please disconnect power and label the address on the hardware!", baud)

    def restore_factory(self) -> None:
        """
        restore factory settings
        :return:
        """
        if "Yes" != input("Restoring factory settings!\n"
                          "Type Yes for confirmation:\n"):
            self.logger.warning("Restore factory settings cancelled.")
            return
        self._execute_valve(bytearray([self.address, 0xff, 0xff, 0xee, 0xbb, 0xaa, 0x00, 0x00, 0x00, 0x00]))
        self.logger.warning("Restored factory settings.")

    def check_motor(self) -> int:
        """
        Check motor status
        :return: status code
        """
        try:
            rtn = self._execute_valve(bytearray([self.address, 0x4a, 0x00, 0x00]))
            status = RunzeMotorEventType(rtn[2])
            self.logger.info("Motor status %s", status.name)
            return status.value
        except ValueError:
            raise RunzeValveError("Unknown motor status")

    def get_current_port(self) -> int:
        """
        Get the current port of the valve
        :return: current port
        """
        rtn = self._execute_valve(bytearray([self.address, 0x3e, 0x00, 0x00]))
        self._port = int(rtn[3])
        self.logger.debug(
            "Valve on %s address %s reports current port %s",
            self.com_port,
            self.address,
            self._port,
        )
        return self._port

    def set_current_port(self, port: int, direction_CW: bool = None, confirmation: bool = True):
        """
        Set the current port of the valve through closest path
        :param port: target port to move to
        :param direction_CW: None to take the closest path, True to move clock wise, False to counter clock wise
        :param confirmation: True to enable confirmation moving to correct port
        :return:
        """
        if not 1 <= port <= self.num_port:
            raise ValueError(f"Port number out of range, default [1, {self.num_port}], set to {port}")
        current_port = getattr(self, "_port", None)
        if current_port is None:
            try:
                current_port = self.get_current_port()
            except Exception:
                current_port = None
        self.logger.info(
            "Requesting Runze valve move on %s address %s from %s to %s (direction_CW=%s, confirmation=%s)",
            self.com_port,
            self.address,
            current_port,
            port,
            direction_CW,
            confirmation,
        )
        if direction_CW is None:
            rtn = self._execute_valve(bytearray([self.address,0x44, port, 0x00]))
        else:
            if direction_CW:
                through_port = 1 if port == self.num_port else port + 1
            else:
                through_port = self.num_port if port == 1 else port - 1
            self.logger.debug(
                "Runze valve on %s address %s moving via through_port=%s to reach port %s",
                self.com_port,
                self.address,
                through_port,
                port,
            )
            rtn = self._execute_valve(bytearray([self.address, 0xa4, through_port, port]))
        if rtn[2] != 0xfe:
            raise RunzeValveError("Set port failed, did not return 'Task being executed'")
        if confirmation:
            time.sleep(0.5)
            current_port = self.get_current_port()
            if current_port != port:
                raise RunzeValveError(f"Set port failed: Expected {port}, got {current_port}")
            self.logger.info(
                "Runze valve on %s address %s moved to port %s",
                self.com_port,
                self.address,
                current_port,
            )
        else:
            self._port = port
            self.logger.info(
                "Runze valve on %s address %s command accepted for port %s without confirmation",
                self.com_port,
                self.address,
                port,
            )

    def reset(self, origin: bool = False) -> None:
        """
        Reset the valve to home position
        :param origin: reset to original home of the encoder
        :return:
        """
        self.logger.info(
            "Resetting Runze valve on %s address %s (origin=%s)",
            self.com_port,
            self.address,
            origin,
        )
        if origin:
            self._execute_valve(bytearray([self.address, 0x4f, 0x00, 0x00]))
        else:
            self._execute_valve(bytearray([self.address, 0x45, 0x00, 0x00]))
        self._port=1
        self.logger.info(
            "Runze valve on %s address %s reset to port %s",
            self.com_port,
            self.address,
            self._port,
        )

    def set_between_ports(self, port) -> None:
        """
        Move the
        :param port: 0.5 position AFTER port (e.g. if port == 1, valve move to between 1 and 2)
        :return:
        """
        if not 1 <= port <= self.num_port:
            raise ValueError(f"Port number out of range, default [1, {self.num_port}], set to {port}")
        port_adjacent = 1 if port == self.num_port else port + 1
        self.logger.info(
            "Moving Runze valve on %s address %s between ports %s and %s",
            self.com_port,
            self.address,
            port,
            port_adjacent,
        )
        self._execute_valve(bytearray([self.address, 0xb4, port, port_adjacent]))

    def _home_valve(self) -> None:
        """
        Home valve, same function name as of Valco
        :return:
        """
        self.reset(origin=False)

    @property
    def port(self) -> int:
        """
        concrete method to report the port of valve
        """
        return self.get_current_port()

    @port.setter
    def port(self, port: int):
        """
        concrete method to set the port of valve
        """
        self.set_current_port(port=port)